var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
// mongoose is not in package.json to avoid compatibility issues with npm package user
const mongoose = require("mongoose");
const { PLUGIN_OPTIONS } = require("./constants");
const { getModelsQueue } = require("./modelsQueue");
function sortEncryptedFields(schema, pluginOptions) {
    const sortEncryptedFieldsOptions = Object.assign(Object.assign({}, PLUGIN_OPTIONS), pluginOptions);
    const { redisQueueClientOptions, ignoreCases, noOfCharsToIncreaseOnSaturation } = sortEncryptedFieldsOptions;
    const sortFields = {};
    const decrypters = {};
    for (const [fieldName, field] of Object.entries(schema.paths)) {
        if (!field.options.sortFieldName)
            continue;
        if (!sortFields[fieldName])
            sortFields[fieldName] = field.options.sortFieldName;
        if (!decrypters[fieldName])
            decrypters[fieldName] = field.options.get;
        schema.add({
            [field.options.sortFieldName]: {
                type: String,
                default: null,
            },
        });
    }
    sortEncryptedFieldsOptions.sortFields = sortFields;
    sortEncryptedFieldsOptions.decrypters = decrypters;
    const modelsQueue = getModelsQueue(redisQueueClientOptions);
    sortEncryptedFieldsOptions.modelsQueue = modelsQueue;
    schema.options.sortEncryptedFieldsOptions = sortEncryptedFieldsOptions;
    schema.post("save", function save(doc, next) {
        return __awaiter(this, void 0, void 0, function* () {
            for (const [fieldName, sortFieldName] of Object.entries(sortFields)) {
                yield modelsQueue.addJob(`${this.constructor.modelName}${sortFieldName}`, {
                    objectId: doc._id,
                    fieldName,
                    fieldValue: doc[fieldName],
                    sortFieldName,
                    ignoreCases,
                    noOfCharsToIncreaseOnSaturation,
                });
            }
            next();
        });
    });
    schema.pre("updateOne", function preUpdateOne(res, next) {
        return __awaiter(this, void 0, void 0, function* () {
            const update = this.getUpdate();
            for (const fieldName in sortFields) {
                const sortFieldName = sortFields[fieldName];
                if (update.$set && update.$set[sortFieldName]) {
                    // Bypass middleware internal call for updating any sortFieldName field
                    break;
                }
                if (update.$set && update.$set[fieldName]) {
                    update.$set[sortFieldName] = null;
                }
            }
            next();
        });
    });
    schema.post("updateOne", function postUpdateOne(res, next) {
        return __awaiter(this, void 0, void 0, function* () {
            const update = this.getUpdate();
            for (const fieldName in sortFields) {
                const sortFieldName = sortFields[fieldName];
                if (update.$set && update.$set[sortFieldName]) {
                    // Bypass middleware internal call for updating any sortFieldName field
                    break;
                }
                if (update.$set && update.$set[fieldName]) {
                    const document = yield this.model.findOne(this.getFilter(), { _id: 1, [fieldName]: 1 }).exec();
                    if (document) {
                        const fieldValue = document[fieldName];
                        yield modelsQueue.addJob(`${this.modelName.model}${sortFieldName}`, {
                            objectId: document._id,
                            fieldName,
                            fieldValue,
                            sortFieldName,
                            ignoreCases,
                            noOfCharsToIncreaseOnSaturation,
                        });
                    }
                }
            }
            next();
        });
    });
    schema.pre("updateMany", function preUpdateMany(res, next) {
        return __awaiter(this, void 0, void 0, function* () {
            const update = this.getUpdate();
            for (const fieldName in sortFields) {
                const sortFieldName = sortFields[fieldName];
                if (update.$set && update.$set[sortFieldName]) {
                    // Bypass middleware internal call for updating any sortFieldName field
                    break;
                }
                if (update.$set && update.$set[fieldName]) {
                    update.$set[sortFieldName] = null;
                }
            }
            next();
        });
    });
    schema.post("updateMany", function postUpdateMany(res, next) {
        return __awaiter(this, void 0, void 0, function* () {
            const update = this.getUpdate();
            for (const fieldName in sortFields) {
                const sortFieldName = sortFields[fieldName];
                if (update.$set && update.$set[sortFieldName]) {
                    // Bypass middleware internal call for updating any sortFieldName field
                    break;
                }
                if (update.$set && update.$set[fieldName]) {
                    const fieldValue = update.$set[fieldName];
                    const documents = yield this.model.find(this.getFilter(), { _id: 1 }).exec();
                    if (documents && documents.length > 0) {
                        for (let i = 0; i < documents.length; i += 1) {
                            yield modelsQueue.addJob(`${this.model.modelName}${sortFieldName}`, {
                                objectId: documents[i]._id,
                                fieldName,
                                fieldValue,
                                sortFieldName,
                                ignoreCases,
                                noOfCharsToIncreaseOnSaturation,
                            });
                        }
                    }
                }
            }
            next();
        });
    });
}
function getModelWithSortEncryptedFieldsPlugin(documentName, schema, pluginOptions) {
    schema.plugin(sortEncryptedFields, pluginOptions);
    const { ignoreCases, noOfCharsToIncreaseOnSaturation, sortFields, modelsQueue, revaluateAllThreshold, revaluateAllCountThreshold } = schema.options.sortEncryptedFieldsOptions;
    const model = mongoose.model(documentName, schema);
    for (const fieldName in sortFields) {
        modelsQueue.registerGroup(`${model.modelName}${sortFields[fieldName]}`, model);
    }
    model
        .find({})
        .count()
        .exec()
        .then((noOfTotalDocuments) => __awaiter(this, void 0, void 0, function* () {
        for (const fieldName in sortFields) {
            const sortFieldName = sortFields[fieldName];
            const noOfDocumentsWithoutSortId = yield model
                .find({ $or: [{ [sortFieldName]: null }, { [sortFieldName]: { $exists: false } }] })
                .count()
                .exec();
            if (noOfTotalDocuments <= revaluateAllCountThreshold || noOfDocumentsWithoutSortId / noOfTotalDocuments > revaluateAllThreshold) {
                yield modelsQueue.removeAllJobs(`${model.modelName}${sortFieldName}`);
                yield modelsQueue.addJob(`${model.modelName}${sortFieldName}`, {
                    generateSortIdForAllDocuments: true,
                    fieldName,
                    sortFieldName,
                    ignoreCases,
                });
            }
            else {
                const documents = yield model
                    .find({ $or: [{ [sortFieldName]: null }, { [sortFieldName]: { $exists: false } }] }, { _id: 1, [fieldName]: 1 })
                    .exec();
                if (documents && documents.length > 0) {
                    for (let i = 0; i < documents.length; i += 1) {
                        const fieldValue = documents[i][fieldName];
                        yield modelsQueue.addJob(`${model.modelName}${sortFieldName}`, {
                            objectId: documents[i]._id,
                            fieldName,
                            fieldValue,
                            sortFieldName,
                            ignoreCases,
                            noOfCharsToIncreaseOnSaturation,
                        });
                    }
                }
            }
        }
    }));
    return model;
}
module.exports = { getModelWithSortEncryptedFieldsPlugin };
